@@ -1,5 +1,5 @@
-/* jdupes (C) 2015-2017 Jody Bruchon <jody@jodybruchon.com>
-   Derived from fdupes (C) 1999-2017 Adrian Lopez
+/* jdupes (C) 2015-2018 Jody Bruchon <jody@jodybruchon.com>
+   Derived from fdupes (C) 1999-2018 Adrian Lopez
 
    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation files
@@ -41,7 +41,14 @@
 #include <sys/time.h>
 #include "jdupes.h"
 #include "string_malloc.h"
-#include "jody_hash.h"
+#if defined USE_HASH_JODYHASH
+ #define JODY_HASH_NOCOMPAT
+ #include "jody_hash.h"
+#elif defined USE_HASH_XXHASH64
+ #include "xxhash.h"
+#else
+ #error No USE_HASH is defined
+#endif
 #include "jody_sort.h"
 #include "jody_win_unicode.h"
 #include "jody_cacheinfo.h"
@@ -89,7 +96,7 @@ int err_mode = _O_TEXT;
 #endif
 
 /* Behavior modification flags */
-uint_fast32_t flags = 0;
+uint_fast32_t flags = 0, p_flags = 0;
 
 static const char *program_name;
 
@@ -101,14 +108,16 @@ struct stat s;
 #endif
 
 /* Larger chunk size makes large files process faster but uses more RAM */
+#define MIN_CHUNK_SIZE 4096
+#define MAX_CHUNK_SIZE 16777216
 #ifndef CHUNK_SIZE
- #define CHUNK_SIZE 32768
+ #define CHUNK_SIZE 65536
 #endif
 #ifndef PARTIAL_HASH_SIZE
  #define PARTIAL_HASH_SIZE 4096
 #endif
 
-static size_t auto_chunk_size;
+static size_t auto_chunk_size = CHUNK_SIZE;
 
 /* Maximum path buffer size to use; must be large enough for a path plus
  * any work that might be done to the array it's stored in. PATH_MAX is
@@ -186,26 +195,31 @@ static const char *extensions[] = {
     #ifdef SMA_PAGE_SIZE
     "smapage",
     #endif
+    #ifdef USE_JODY_HASH
+    #if JODY_HASH_WIDTH == 64
+    "jodyhash64",
+    #endif
     #if JODY_HASH_WIDTH == 32
-    "hash32",
+    "jodyhash32",
     #endif
     #if JODY_HASH_WIDTH == 16
-    "hash16",
+    "jodyhash16",
+    #endif
+    #endif /* USE_JODY_HASH */
+    #ifdef USE_HASH_XXHASH64
+    "xxhash64",
     #endif
     #ifdef NO_PERMS
     "noperm",
     #endif
+    #ifdef NO_HARDLINKS
+    "nohardlink",
+    #endif
     #ifdef NO_SYMLINKS
     "nosymlink",
     #endif
-    #ifdef USE_TREE_REBALANCE
-    "rebal",
-    #endif
-    #ifdef CONSIDER_IMBALANCE
-    "ci",
-    #endif
-    #ifdef BALANCE_THRESHOLD
-    "bt",
+    #ifdef NO_USER_ORDER
+    "nouserorder",
     #endif
     NULL
 };
@@ -556,7 +570,8 @@ bad_size_suffix:
 
 
 extern int getdirstats(const char * const restrict name,
-        jdupes_ino_t * const restrict inode, dev_t * const restrict dev)
+        jdupes_ino_t * const restrict inode, dev_t * const restrict dev,
+        jdupes_mode_t * const restrict mode)
 {
   if (name == NULL || inode == NULL || dev == NULL) nullptr("getdirstats");
   LOUD(fprintf(stderr, "getdirstats('%s', %p, %p)\n", name, (void *)inode, (void *)dev);)
@@ -565,11 +580,13 @@ extern int getdirstats(const char * const restrict name,
   if (win_stat(name, &ws) != 0) return -1;
   *inode = ws.inode;
   *dev = ws.device;
+  *mode = ws.mode;
   if (!S_ISDIR(ws.mode)) return 1;
 #else
   if (stat(name, &s) != 0) return -1;
   *inode = s.st_ino;
   *dev = s.st_dev;
+  *mode = s.st_mode;
   if (!S_ISDIR(s.st_mode)) return 1;
 #endif /* ON_WINDOWS */
   return 0;
@@ -588,11 +605,13 @@ extern int check_conditions(const file_t * const restrict file1, const file_t *
 
   LOUD(fprintf(stderr, "check_conditions('%s', '%s')\n", file1->d_name, file2->d_name);)
 
+#ifndef NO_USER_ORDER
   /* Exclude based on -I/--isolate */
   if (ISFLAG(flags, F_ISOLATE) && (file1->user_order == file2->user_order)) {
     LOUD(fprintf(stderr, "check_conditions: files ignored: parameter isolation\n"));
     return -1;
   }
+#endif /* NO_USER_ORDER */
 
   /* Exclude based on -1/--one-file-system */
   if (ISFLAG(flags, F_ONEFS) && (file1->device != file2->device)) {
@@ -647,7 +666,7 @@ extern int check_conditions(const file_t * const restrict file1, const file_t *
 static int check_singlefile(file_t * const restrict newfile)
 {
   static char tempname[PATHBUF_SIZE * 2];
-  static char *tp = tempname;
+  char * restrict tp = tempname;
   int excluded;
 
   if (newfile == NULL) nullptr("check_singlefile()");
@@ -656,6 +675,7 @@ static int check_singlefile(file_t * const restrict newfile)
 
   /* Exclude hidden files if requested */
   if (ISFLAG(flags, F_EXCLUDEHIDDEN)) {
+    if (newfile->d_name == NULL) nullptr("check_singlefile newfile->d_name");
     strcpy(tp, newfile->d_name);
     tp = basename(tp);
     if (tp[0] == '.' && strcmp(tp, ".") && strcmp(tp, "..")) {
@@ -716,13 +736,20 @@ static int check_singlefile(file_t * const restrict newfile)
 static file_t *init_newfile(const size_t len, file_t * restrict * const restrict filelistp)
 {
   file_t * const restrict newfile = (file_t *)string_malloc(sizeof(file_t));
+
   if (!newfile) oom("init_newfile() file structure");
+  if (!filelistp) nullptr("init_newfile() filelistp");
+
+  LOUD(fprintf(stderr, "init_newfile(len %lu, filelistp %p)\n", len, filelistp));
+
   memset(newfile, 0, sizeof(file_t));
   newfile->d_name = (char *)string_malloc(len);
   if (!newfile->d_name) oom("init_newfile() filename");
 
   newfile->next = *filelistp;
+#ifndef NO_USER_ORDER
   newfile->user_order = user_item_count;
+#endif
   newfile->size = -1;
   newfile->duplicates = NULL;
   return newfile;
@@ -788,6 +815,7 @@ static void grokdir(const char * const restrict dir,
   int i, single = 0;
   jdupes_ino_t inode, n_inode;
   dev_t device, n_device;
+  jdupes_mode_t mode;
 #ifdef UNICODE
   WIN32_FIND_DATA ffd;
   HANDLE hFind = INVALID_HANDLE_VALUE;
@@ -800,7 +828,7 @@ static void grokdir(const char * const restrict dir,
   LOUD(fprintf(stderr, "grokdir: scanning '%s' (order %d, recurse %d)\n", dir, user_item_count, recurse));
 
   /* Double traversal prevention tree */
-  i = getdirstats(dir, &inode, &device);
+  i = getdirstats(dir, &inode, &device, &mode);
   if (i < 0) goto error_travdone;
 
   if (travdone_head == NULL) {
@@ -811,7 +839,7 @@ static void grokdir(const char * const restrict dir,
     while (1) {
       if (traverse == NULL) nullptr("grokdir() traverse");
       /* Don't re-traverse directories we've already seen */
-      if (inode == traverse->inode && device == traverse->device) {
+      if (S_ISDIR(mode) && inode == traverse->inode && device == traverse->device) {
         LOUD(fprintf(stderr, "already seen item '%s', skipping\n", dir);)
         return;
       } else if (inode > traverse->inode || (inode == traverse->inode && device > traverse->device)) {
@@ -927,7 +955,7 @@ static void grokdir(const char * const restrict dir,
       if (recurse) {
         /* --one-file-system */
         if (ISFLAG(flags, F_ONEFS)
-            && (getdirstats(newfile->d_name, &n_inode, &n_device) == 0)
+            && (getdirstats(newfile->d_name, &n_inode, &n_device, &mode) == 0)
             && (device != n_device)) {
           LOUD(fprintf(stderr, "grokdir: directory: not recursing (--one-file-system)\n"));
           string_free(newfile->d_name);
@@ -1007,19 +1035,28 @@ error_overflow:
 
 
 /* Use Jody Bruchon's hash function on part or all of a file */
-static hash_t *get_filehash(const file_t * const restrict checkfile,
+static jdupes_hash_t *get_filehash(const file_t * const restrict checkfile,
                 const size_t max_read)
 {
   off_t fsize;
   /* This is an array because we return a pointer to it */
-  static hash_t hash[1];
-  static hash_t chunk[(CHUNK_SIZE / sizeof(hash_t))];
+  static jdupes_hash_t hash[1];
+  static jdupes_hash_t *chunk = NULL;
   FILE *file;
   int check = 0;
+#ifdef USE_HASH_XXHASH64
+  XXH64_state_t *xxhstate;
+#endif
 
   if (checkfile == NULL || checkfile->d_name == NULL) nullptr("get_filehash()");
   LOUD(fprintf(stderr, "get_filehash('%s', %" PRIdMAX ")\n", checkfile->d_name, (intmax_t)max_read);)
 
+  /* Allocate on first use */
+  if (chunk == NULL) {
+    chunk = (jdupes_hash_t *)string_malloc(auto_chunk_size);
+    if (!chunk) oom("get_filehash() chunk");
+  }
+
   /* Get the file size. If we can't read it, bail out early */
   if (checkfile->size == -1) {
     LOUD(fprintf(stderr, "get_filehash: not hashing because stat() info is bad\n"));
@@ -1036,8 +1073,7 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
    * If we already hashed the first chunk of this file, we don't want to
    * wastefully read and hash it again, so skip the first chunk and use
    * the computed hash for that chunk as our starting point.
-   *
-   * WARNING: We assume max_read is NEVER less than CHUNK_SIZE here! */
+   */
 
   *hash = 0;
   if (ISFLAG(checkfile->flags, F_HASH_PARTIAL)) {
@@ -1069,6 +1105,13 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
     }
     fsize -= PARTIAL_HASH_SIZE;
   }
+
+#ifdef USE_HASH_XXHASH64
+  xxhstate = XXH64_createState();
+  if (xxhstate == NULL) nullptr("xxhstate");
+  XXH64_reset(xxhstate, 0);
+#endif
+
   /* Read the file in CHUNK_SIZE chunks until we've read it all. */
   while (fsize > 0) {
     size_t bytes_to_read;
@@ -1081,7 +1124,12 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
       return NULL;
     }
 
+#if defined USE_HASH_JODYHASH
     *hash = jody_block_hash(chunk, *hash, bytes_to_read);
+#elif defined USE_HASH_XXHASH64
+    XXH64_update(xxhstate, chunk, bytes_to_read);
+#endif
+
     if ((off_t)bytes_to_read > fsize) break;
     else fsize -= (off_t)bytes_to_read;
 
@@ -1096,6 +1144,10 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
 
   fclose(file);
 
+#ifdef USE_HASH_XXHASH64
+  *hash = XXH64_digest(xxhstate);
+  XXH64_freeState(xxhstate);
+#endif
   LOUD(fprintf(stderr, "get_filehash: returning hash: 0x%016jx\n", (uintmax_t)*hash));
   return hash;
 }
@@ -1115,49 +1167,7 @@ static inline void registerfile(filetree_t * restrict * const restrict nodeptr,
   branch->file = file;
   branch->left = NULL;
   branch->right = NULL;
-#ifdef USE_TREE_REBALANCE
-  branch->left_weight = 0;
-  branch->right_weight = 0;
 
-  /* Attach the new node to the requested branch and the parent */
-  switch (d) {
-    case LEFT:
-      branch->parent = *nodeptr;
-      (*nodeptr)->left = branch;
-      (*nodeptr)->left_weight++;
-      break;
-    case RIGHT:
-      branch->parent = *nodeptr;
-      (*nodeptr)->right = branch;
-      (*nodeptr)->right_weight++;
-      break;
-    case NONE:
-      /* For the root of the tree only */
-      branch->parent = NULL;
-      *nodeptr = branch;
-      break;
-    default:
-      /* This should never ever happen */
-      fprintf(stderr, "\ninternal error: invalid direction for registerfile(), report this\n");
-      string_malloc_destroy();
-      exit(EXIT_FAILURE);
-      break;
-  }
-
-  /* Propagate weights up the tree */
-  while (branch->parent != NULL) {
-    filetree_t * restrict up;
-
-    up = branch->parent;
-    if (up->left == branch) up->left_weight++;
-    else if (up->right == branch) up->right_weight++;
-    else {
-      fprintf(stderr, "\nInternal error: file tree linkage is broken\n");
-      exit(EXIT_FAILURE);
-    }
-    branch = up;
-  }
-#else /* USE_TREE_REBALANCE */
   /* Attach the new node to the requested branch */
   switch (d) {
     case LEFT:
@@ -1178,109 +1188,10 @@ static inline void registerfile(filetree_t * restrict * const restrict nodeptr,
       break;
   }
 
-#endif /* USE_TREE_REBALANCE */
-
   return;
 }
 
 
-/* Experimental tree rebalance code. This slows things down in testing
- * but may be more useful in the future. Pass -DUSE_TREE_REBALANCE
- * to try it. */
-#ifdef USE_TREE_REBALANCE
-
-/* How much difference to ignore when considering a rebalance */
-#ifndef BALANCE_THRESHOLD
-#define BALANCE_THRESHOLD 4
-#endif
-
-/* Rebalance the file tree to reduce search depth */
-static inline void rebalance_tree(filetree_t * const tree)
-{
-  filetree_t * restrict promote;
-  filetree_t * restrict demote;
-  int difference, direction;
-#ifdef CONSIDER_IMBALANCE
-  int l, r, imbalance;
-#endif
-
-  if (!tree) return;
-
-  /* Rebalance all children first */
-  if (tree->left_weight > BALANCE_THRESHOLD) rebalance_tree(tree->left);
-  if (tree->right_weight > BALANCE_THRESHOLD) rebalance_tree(tree->right);
-
-  /* If weights are within a certain threshold, do nothing */
-  direction = tree->right_weight - tree->left_weight;
-  difference = direction;
-  if (difference < 0) difference = -difference;
-  if (difference <= BALANCE_THRESHOLD) return;
-
-  /* Determine if a tree rotation will help, and do it if so */
-  if (direction > 0) {
-#ifdef CONSIDER_IMBALANCE
-    l = tree->right->left_weight + tree->right_weight;
-    r = tree->right->right_weight;
-    imbalance = l - r;
-    if (imbalance < 0) imbalance = -imbalance;
-    /* Don't rotate if imbalance will increase */
-    if (imbalance >= difference) return;
-#endif /* CONSIDER_IMBALANCE */
-
-    /* Rotate the right node up one level */
-    promote = tree->right;
-    demote = tree;
-    /* Attach new parent's left tree to old parent */
-    demote->right = promote->left;
-    demote->right_weight = promote->left_weight;
-    /* Attach old parent to new parent */
-    promote->left = demote;
-    promote->left_weight = demote->left_weight + demote->right_weight + 1;
-    /* Reconnect parent linkages */
-    promote->parent = demote->parent;
-    if (demote->right) demote->right->parent = demote;
-    demote->parent = promote;
-    if (promote->parent == NULL) checktree = promote;
-    else if (promote->parent->left == demote) promote->parent->left = promote;
-    else promote->parent->right = promote;
-    return;
-  } else if (direction < 0) {
-#ifdef CONSIDER_IMBALANCE
-    r = tree->left->right_weight + tree->left_weight;
-    l = tree->left->left_weight;
-    imbalance = r - l;
-    if (imbalance < 0) imbalance = -imbalance;
-    /* Don't rotate if imbalance will increase */
-    if (imbalance >= difference) return;
-#endif /* CONSIDER_IMBALANCE */
-
-    /* Rotate the left node up one level */
-    promote = tree->left;
-    demote = tree;
-    /* Attach new parent's right tree to old parent */
-    demote->left = promote->right;
-    demote->left_weight = promote->right_weight;
-    /* Attach old parent to new parent */
-    promote->right = demote;
-    promote->right_weight = demote->right_weight + demote->left_weight + 1;
-    /* Reconnect parent linkages */
-    promote->parent = demote->parent;
-    if (demote->left) demote->left->parent = demote;
-    demote->parent = promote;
-    if (promote->parent == NULL) checktree = promote;
-    else if (promote->parent->left == demote) promote->parent->left = promote;
-    else promote->parent->right = promote;
-    return;
-
-  }
-
-  /* Fall through */
-  return;
-}
-
-#endif /* USE_TREE_REBALANCE */
-
-
 #ifdef TREE_DEPTH_STATS
 #define TREE_DEPTH_UPDATE_MAX() { if (max_depth < tree_depth) max_depth = tree_depth; tree_depth = 0; }
 #else
@@ -1292,7 +1203,7 @@ static inline void rebalance_tree(filetree_t * const tree)
 static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict file)
 {
   int cmpresult = 0;
-  const hash_t * restrict filehash;
+  const jdupes_hash_t * restrict filehash;
 
   if (tree == NULL || file == NULL || tree->file == NULL || tree->file->d_name == NULL || file->d_name == NULL) nullptr("checkmatch()");
   LOUD(fprintf(stderr, "checkmatch ('%s', '%s')\n", tree->file->d_name, file->d_name));
@@ -1317,7 +1228,10 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
     default: break;
   }
 
-  /* If preliminary matching succeeded, move to full file checks */
+  /* Print pre-check (early) match candidates if requested */
+  if (ISFLAG(p_flags, P_EARLYMATCH)) printf("Early match check passed:\n   %s\n   %s\n\n", file->d_name, tree->file->d_name);
+
+  /* If preliminary matching succeeded, do main file data checks */
   if (cmpresult == 0) {
     LOUD(fprintf(stderr, "checkmatch: starting file data comparisons\n"));
     /* Attempt to exclude files quickly with partial file hashing */
@@ -1373,6 +1287,9 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
         SETFLAG(file->flags, F_HASH_FULL);
       }
 
+      /* Print partial hash matching pairs if requested */
+      if (ISFLAG(p_flags, P_PARTIAL)) printf("Partial hashes match:\n   %s\n   %s\n\n", file->d_name, tree->file->d_name);
+
       /* Full file hash comparison */
       cmpresult = HASH_COMPARE(file->filehash, tree->file->filehash);
       LOUD(if (!cmpresult) fprintf(stderr, "checkmatch: full hashes match\n"));
@@ -1410,6 +1327,7 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
     DBG(partial_to_full++;)
     TREE_DEPTH_UPDATE_MAX();
     LOUD(fprintf(stderr, "checkmatch: files appear to match based on hashes\n"));
+    if (ISFLAG(p_flags, P_FULLHASH)) printf("Full hashes match:\n   %s\n   %s\n\n", file->d_name, tree->file->d_name);
     return &tree->file;
   }
   /* Fall through - should never be reached */
@@ -1421,7 +1339,7 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
    same signature. Unlikely, but better safe than sorry. */
 static inline int confirmmatch(FILE * const restrict file1, FILE * const restrict file2, off_t size)
 {
-  static char c1[CHUNK_SIZE], c2[CHUNK_SIZE];
+  static char *c1 = NULL, *c2 = NULL;
   size_t r1, r2;
   off_t bytes = 0;
   int check = 0;
@@ -1429,6 +1347,13 @@ static inline int confirmmatch(FILE * const restrict file1, FILE * const restric
   if (file1 == NULL || file2 == NULL) nullptr("confirmmatch()");
   LOUD(fprintf(stderr, "confirmmatch running\n"));
 
+  /* Allocate on first use; OOM if either is ever NULLed */
+  if (!c1) {
+    c1 = (char *)string_malloc(auto_chunk_size);
+    c2 = (char *)string_malloc(auto_chunk_size);
+  }
+  if (!c1 || !c2) oom("confirmmatch() c1/c2");
+
   fseek(file1, 0, SEEK_SET);
   fseek(file2, 0, SEEK_SET);
 
@@ -1483,6 +1408,7 @@ extern unsigned int get_max_dupes(const file_t *files, unsigned int * const rest
 }
 
 
+#ifndef NO_USER_ORDER
 static int sort_pairs_by_param_order(file_t *f1, file_t *f2)
 {
   if (!ISFLAG(flags, F_USEPARAMORDER)) return 0;
@@ -1491,14 +1417,17 @@ static int sort_pairs_by_param_order(file_t *f1, file_t *f2)
   if (f1->user_order > f2->user_order) return sort_direction;
   return 0;
 }
+#endif
 
 
 static int sort_pairs_by_mtime(file_t *f1, file_t *f2)
 {
   if (f1 == NULL || f2 == NULL) nullptr("sort_pairs_by_mtime()");
-  int po = sort_pairs_by_param_order(f1, f2);
 
+#ifndef NO_USER_ORDER
+  int po = sort_pairs_by_param_order(f1, f2);
   if (po != 0) return po;
+#endif /* NO_USER_ORDER */
 
   if (f1->mtime < f2->mtime) return -sort_direction;
   else if (f1->mtime > f2->mtime) return sort_direction;
@@ -1510,9 +1439,11 @@ static int sort_pairs_by_mtime(file_t *f1, file_t *f2)
 static int sort_pairs_by_filename(file_t *f1, file_t *f2)
 {
   if (f1 == NULL || f2 == NULL) nullptr("sort_pairs_by_filename()");
-  int po = sort_pairs_by_param_order(f1, f2);
 
+#ifndef NO_USER_ORDER
+  int po = sort_pairs_by_param_order(f1, f2);
   if (po != 0) return po;
+#endif /* NO_USER_ORDER */
 
   return numeric_sort(f1->d_name, f2->d_name, sort_direction);
 }
@@ -1573,8 +1504,9 @@ static inline void help_text(void)
   printf(" -1 --one-file-system \tdo not match files on different filesystems/devices\n");
   printf(" -A --nohidden    \texclude hidden files from consideration\n");
 #ifdef ENABLE_BTRFS
-  printf(" -B --dedupe      \tSend matches to btrfs for block-level deduplication\n");
+  printf(" -B --dedupe      \tsend matches to btrfs for block-level deduplication\n");
 #endif
+  printf(" -C --chunksize=# \toverride I/O chunk size (min %d, max %d)\n", MIN_CHUNK_SIZE, MAX_CHUNK_SIZE);
   printf(" -d --delete      \tprompt user for files to preserve and delete all\n");
   printf("                  \tothers; important: under particular circumstances,\n");
   printf("                  \tdata may be lost when using this option together\n");
@@ -1591,7 +1523,9 @@ static inline void help_text(void)
   printf("                  \tlinked files are treated as non-duplicates for safety\n");
 #endif
   printf(" -i --reverse     \treverse (invert) the match sort order\n");
+#ifndef NO_USER_ORDER
   printf(" -I --isolate     \tfiles in the same specified directory won't match\n");
+#endif
 #ifndef NO_SYMLINKS
   printf(" -l --linksoft    \tmake relative symlinks for duplicates w/o prompting\n");
 #endif
@@ -1607,12 +1541,15 @@ static inline void help_text(void)
   printf("                  \teach set of duplicates and delete the rest without\n");
   printf("                  \tprompting the user\n");
   printf(" -o --order=BY    \tselect sort order for output, linking and deleting; by\n");
+#ifndef NO_USER_ORDER
   printf(" -O --paramorder  \tParameter order is more important than selected -O sort\n");
   printf("                  \tmtime (BY=time) or filename (BY=name, the default)\n");
+#endif
 #ifndef NO_PERMS
   printf(" -p --permissions \tdon't consider files with different owner/group or\n");
   printf("                  \tpermission bits as duplicates\n");
 #endif
+  printf(" -P --print=type  \tprint extra info (partial, early, fullhash)\n");
   printf(" -Q --quick       \tskip byte-for-byte confirmation for quick matching\n");
   printf("                  \tWARNING: -Q can result in data loss! Be very careful!\n");
   printf(" -r --recurse     \tfor every directory, process its subdirectories too\n");
@@ -1632,7 +1569,7 @@ static inline void help_text(void)
   printf(" -x --xsize=SIZE  \texclude files of size < SIZE bytes from consideration\n");
   printf("    --xsize=+SIZE \t'+' specified before SIZE, exclude size > SIZE\n");
   printf(" -X --exclude=spec:info\texclude files based on specified criteria\n");
-  printf("                  \tspecs: dir size+-=\n");
+  printf("                  \tspecs: size+-=\n");
   printf("                  \tExclusions are cumulative: -X dir:abc -X dir:efg\n");
   printf(" -z --zeromatch   \tconsider zero-length files to be duplicates\n");
   printf(" -Z --softabort   \tIf the user aborts (i.e. CTRL-C) act on matches so far\n");
@@ -1649,7 +1586,6 @@ int wmain(int argc, wchar_t **wargv)
 int main(int argc, char **argv)
 #endif
 {
-  static struct proc_cacheinfo pci;
   static file_t *files = NULL;
   static file_t *curfile;
   static char **oldargv;
@@ -1658,6 +1594,10 @@ int main(int argc, char **argv)
   static int opt;
   static int pm = 1;
   static ordertype_t ordertype = ORDER_NAME;
+  static long manual_chunk_size = 0;
+#ifndef ON_WINDOWS
+  static struct proc_cacheinfo pci;
+#endif
 
 #ifndef OMIT_GETOPT_LONG
   static const struct option long_options[] =
@@ -1666,6 +1606,7 @@ int main(int argc, char **argv)
     { "one-file-system", 0, 0, '1' },
     { "nohidden", 0, 0, 'A' },
     { "dedupe", 0, 0, 'B' },
+    { "chunksize", 1, 0, 'C' },
     { "delete", 0, 0, 'd' },
     { "debug", 0, 0, 'D' },
     { "omitfirst", 0, 0, 'f' },
@@ -1682,6 +1623,7 @@ int main(int argc, char **argv)
     { "order", 1, 0, 'o' },
     { "paramorder", 0, 0, 'O' },
     { "permissions", 0, 0, 'p' },
+    { "print", 0, 0, 'P' },
     { "quiet", 0, 0, 'q' },
     { "quick", 0, 0, 'Q' },
     { "recurse", 0, 0, 'r' },
@@ -1723,22 +1665,24 @@ int main(int argc, char **argv)
   else err_mode = _O_U16TEXT;
 #endif /* UNICODE */
 
+#ifndef ON_WINDOWS
   /* Auto-tune chunk size to be half of L1 data cache if possible */
   get_proc_cacheinfo(&pci);
   if (pci.l1 != 0) auto_chunk_size = (pci.l1 / 2);
   else if (pci.l1d != 0) auto_chunk_size = (pci.l1d / 2);
   /* Must be at least 4096 (4 KiB) and cannot exceed CHUNK_SIZE */
-  if (auto_chunk_size < 4096 || auto_chunk_size > CHUNK_SIZE) auto_chunk_size = CHUNK_SIZE;
+  if (auto_chunk_size < MIN_CHUNK_SIZE || auto_chunk_size > MAX_CHUNK_SIZE) auto_chunk_size = CHUNK_SIZE;
   /* Force to a multiple of 4096 if it isn't already */
   if ((auto_chunk_size & 0x00000fffUL) != 0)
     auto_chunk_size = (auto_chunk_size + 0x00000fffUL) & 0x000ff000;
+#endif /* ON_WINDOWS */
 
   program_name = argv[0];
 
   oldargv = cloneargs(argc, argv);
 
   while ((opt = GETOPT(argc, argv,
-  "@1ABdDfhHiIlLmnNOpqQrRsSvzZo:x:X:"
+  "@1ABC:dDfhHiIlLmnNOpP:qQrRsSvzZo:x:X:"
 #ifndef OMIT_GETOPT_LONG
           , long_options, NULL
 #endif
@@ -1750,6 +1694,15 @@ int main(int argc, char **argv)
     case 'A':
       SETFLAG(flags, F_EXCLUDEHIDDEN);
       break;
+    case 'C':
+      manual_chunk_size = strtol(optarg, NULL, 10) & 0x0ffff000L;  /* Align to 4K sizes */
+      if (manual_chunk_size < MIN_CHUNK_SIZE || manual_chunk_size > MAX_CHUNK_SIZE) {
+        fprintf(stderr, "warning: invalid manual chunk size (must be %d-%d); using defaults\n", MIN_CHUNK_SIZE, MAX_CHUNK_SIZE);
+        LOUD(fprintf(stderr, "Manual chunk size (failed) was apparently '%s' => %ld\n", optarg, manual_chunk_size));
+        manual_chunk_size = 0;
+      } else auto_chunk_size = (size_t)manual_chunk_size;
+      LOUD(fprintf(stderr, "Manual chunk size is %ld\n", manual_chunk_size));
+      break;
     case 'd':
       SETFLAG(flags, F_DELETEFILES);
       break;
@@ -1776,9 +1729,19 @@ int main(int argc, char **argv)
     case 'i':
       SETFLAG(flags, F_REVERSESORT);
       break;
+#ifndef NO_USER_ORDER
     case 'I':
       SETFLAG(flags, F_ISOLATE);
       break;
+    case 'O':
+      SETFLAG(flags, F_USEPARAMORDER);
+      break;
+#else
+    case 'I':
+    case 'O':
+      fprintf(stderr, "warning: -I and -O are disabled and ignored in this build\n");
+      break;
+#endif
     case 'm':
       SETFLAG(flags, F_SUMMARIZEMATCHES);
       break;
@@ -1788,12 +1751,18 @@ int main(int argc, char **argv)
     case 'N':
       SETFLAG(flags, F_NOPROMPT);
       break;
-    case 'O':
-      SETFLAG(flags, F_USEPARAMORDER);
-      break;
     case 'p':
       SETFLAG(flags, F_PERMISSIONS);
       break;
+    case 'P':
+      if (strcmp(optarg, "partial") == 0) SETFLAG(p_flags, P_PARTIAL);
+      else if (strcmp(optarg, "early") == 0) SETFLAG(p_flags, P_EARLYMATCH);
+      else if (strcmp(optarg, "fullhash") == 0) SETFLAG(p_flags, P_FULLHASH);
+      else {
+        fprintf(stderr, "Option '%s' is not valid for -P\n", optarg);
+	exit(EXIT_FAILURE);
+      }
+      break;
     case 'q':
       SETFLAG(flags, F_HIDEPROGRESS);
       break;
@@ -1867,7 +1836,7 @@ int main(int argc, char **argv)
           c++;
         }
       } else printf(" none");
-      printf("\nCopyright (C) 2015-2017 by Jody Bruchon\n");
+      printf("\nCopyright (C) 2015-2018 by Jody Bruchon\n");
       printf("\nPermission is hereby granted, free of charge, to any person\n");
       printf("obtaining a copy of this software and associated documentation files\n");
       printf("(the \"Software\"), to deal in the Software without restriction,\n");
@@ -2005,9 +1974,6 @@ int main(int argc, char **argv)
     static file_t **match = NULL;
     static FILE *file1;
     static FILE *file2;
-#ifdef USE_TREE_REBALANCE
-    static unsigned int depth_threshold = INITIAL_DEPTH_THRESHOLD;
-#endif
 
     if (interrupt) {
       fprintf(stderr, "\nStopping file scan due to user abort\n");
@@ -2021,16 +1987,6 @@ int main(int argc, char **argv)
     if (!checktree) registerfile(&checktree, NONE, curfile);
     else match = checkmatch(checktree, curfile);
 
-#ifdef USE_TREE_REBALANCE
-    /* Rebalance the match tree after a certain number of files processed */
-    if (max_depth > depth_threshold) {
-      rebalance_tree(checktree);
-      max_depth = 0;
-      if (depth_threshold < 512) depth_threshold <<= 1;
-      else depth_threshold += 64;
-    }
-#endif /* USE_TREE_REBALANCE */
-
     /* Byte-for-byte check that a matched pair are actually matched */
     if (match != NULL) {
       /* Quick comparison mode will never run confirmmatch()
@@ -2116,16 +2072,22 @@ skip_file_scan:
   if (ISFLAG(flags, F_DEBUG)) {
     fprintf(stderr, "\n%d partial (+%d small) -> %d full hash -> %d full (%d partial elim) (%d hash%u fail)\n",
         partial_hash, small_file, full_hash, partial_to_full,
-        partial_elim, hash_fail, (unsigned int)sizeof(hash_t)*8);
+        partial_elim, hash_fail, (unsigned int)sizeof(jdupes_hash_t)*8);
     fprintf(stderr, "%" PRIuMAX " total files, %" PRIuMAX " comparisons, branch L %u, R %u, both %u, max tree depth %u\n",
         filecount, comparisons, left_branch, right_branch,
         left_branch + right_branch, max_depth);
     fprintf(stderr, "SMA: allocs %" PRIuMAX ", free %" PRIuMAX " (merge %" PRIuMAX ", repl %" PRIuMAX "), fail %" PRIuMAX ", reuse %" PRIuMAX ", scan %" PRIuMAX ", tails %" PRIuMAX "\n",
         sma_allocs, sma_free_good, sma_free_merged, sma_free_replaced,
-	sma_free_ignored, sma_free_reclaimed,
+        sma_free_ignored, sma_free_reclaimed,
         sma_free_scanned, sma_free_tails);
-    fprintf(stderr, "I/O chunk size: %" PRIuMAX " KiB (%s)\n", (uintmax_t)(auto_chunk_size >> 10),
-        (pci.l1 + pci.l1d) != 0 ? "dynamically sized" : "default size");
+    if (manual_chunk_size > 0) fprintf(stderr, "I/O chunk size: %ld KiB (manually set)\n", manual_chunk_size >> 10);
+    else {
+#ifndef ON_WINDOWS
+      fprintf(stderr, "I/O chunk size: %" PRIuMAX " KiB (%s)\n", (uintmax_t)(auto_chunk_size >> 10), (pci.l1 + pci.l1d) != 0 ? "dynamically sized" : "default size");
+#else
+      fprintf(stderr, "I/O chunk size: %" PRIuMAX " KiB (default size)\n", (uintmax_t)(auto_chunk_size >> 10));
+#endif
+    }
 #ifdef ON_WINDOWS
  #ifndef NO_HARDLINKS
     if (ISFLAG(flags, F_HARDLINKFILES))
